from vcPythonProcessHandler import *
from vcBehaviour import * 
import vcMatrix, vcVector, math
import re

RAD_TO_DEG = math.pi/180.0 
  
app = getApplication() 
sim = getSimulation()
comp = getComponent()

ctr = None
mt = None
tracking_on = False
base_node = None
last_time = -1


class RE(object):
  #Capsulates RE constants
  
  ANY = r'.+'
  ANYQ = r'.*'
  SPC = r'\s+'
  SPCQ = r'\s*'
  EOL = r'\n'
  EOS = r'$'
  INT = r'\s*-?\d+'
  NAME = r'\w+'
  NUM = r'\s*-?\d+\.?\d*[eE]?-?\d*'
  VAR = r'[a-z][a-z0-9]*\[[0-9]+\]|[a-z][a-z0-9]*'
  FLAGS = re.I|re.S


def OnReset():
  global ctr, mt, tracking_on, base_node, last_time
  
  ctr = None
  mt = None
  tracking_on = False
  comp.OnNodeTransform = None
  if base_node:
    try:
      base_node.OnNodeTransform = None
    except:
      pass
  base_node = None
  last_time = -1


def OnStatementAdd(stmt):
  for prop in stmt.Properties:
    if prop.Name in ['Name', 'Tool', 'Process', 'icon']:
      prop.IsVisible = False


def OnStatementGetDescription(stmt):
  base_name = ''
  if stmt.Base:
    base_name = stmt.Base.Name
  return '%s %s' % (stmt.Name, base_name)


def OnStatementExecute(exe, stmt):
  global ctr, mt, tracking_on, base_node, last_time
  
  base = stmt.Base
  if not base:
    return
  ctr = exe.Controller
  
  if 'Wait' in stmt.Name:
    # WaitPart logic
    prop_sensor = stmt.getProperty('Sensor')
    if not prop_sensor.Value:
      return
    sensor_node = prop_sensor.Value
    
    # Try finding component signal related to sensor
    bs = sensor_node.getBehavioursByType(VC_COMPONENTSIGNAL)
    if not bs:
      return
    signal = bs[0]
    
    self_handle = comp.findBehaviour('CTHandler')
    if not self_handle in signal.Connections:
      connections = signal.Connections
      connections.append(self_handle)
      signal.Connections = connections
    
    part = None
    while not part:
      triggerCondition(lambda: getTrigger() == signal)
      part = signal.Value
    base.Node = part
    base.PositionMatrix = vcMatrix.new()
    
  elif 'Track' in stmt.Name:
    # TrackPart logic
    
    # Check that base is set to other node than robot itself
    if not base or not base.Node or base.Node.Component == ctr.Component:
      return
    
    comp.update()
    base.Node.Component.update()
    
    base_node = base.Node
    last_time = -1
    prop_stop_with = stmt.getProperty('StopWith')
    prop_timer = stmt.getProperty('Timer')
    prop_condition = stmt.getProperty('Condition')
    init_exp = format_expression(prop_condition.Value)
    if not (prop_stop_with and prop_timer and prop_condition):
      return
    
    # Get test expression property
    prop_test = comp.getProperty('TestExpression')
    if not prop_test:
      prop_test = comp.createProperty(VC_EXPRESSION, 'TestExpression')
    
    # Create target to track
    mt = ctr.createTarget()
    
    # Reset lag by getting accurate position from previous motion statement
    scope = stmt.ParentScope
    stmt_index = -1
    for i in range(len(scope.Statements)):
      if scope.Statements[i] == stmt:
        stmt_index = i
        break
    while stmt_index > 0:
      stmt_index -= 1
      prev_stmt = scope.Statements[stmt_index]
      if prev_stmt.Type in [VC_STATEMENT_LINMOTION, VC_STATEMENT_PTPMOTION]:
        mt.Target = prev_stmt.Positions[0].PositionInReference
        break
      elif prev_stmt.Type in [VC_STATEMENT_COMMENT, VC_STATEMENT_PRINT, VC_STATEMENT_SETBIN, VC_STATEMENT_SETPROPERTY]:
        continue
      else:
        break
      
    if base:
      mt.BaseName = base.Name
    else:
      mt.BaseName = ''
    tgt = mt.Target
    ctr.moveImmediate(mt)
    
    # Idle with tracking active until timer or conditions breaks
    start_time = sim.SimTime
    comp.OnNodeTransform = track
    base_node.OnNodeTransform = track
    tracking_on = True
    while True:
      if prop_stop_with.Value == 'Timer':
        if sim.SimTime - start_time >= prop_timer.Value:
          break
      elif prop_stop_with.Value == 'Condition':
        exp = format_expression(init_exp, exe)
        prop_test.Value = exp
        if prop_test.CalculatedValue:
          # Wait condition is true, break
          break
      else: #if prop_stop_with.Value == 'Reset':
        break
      
      delay(0.01)
    #endwhile
    
    if prop_stop_with.Value != 'Reset':
      tracking_on = False
      comp.OnNodeTransform = None
      base_node.OnNodeTransform = None
    
  else: #if 'Reset' in stmt.Name:
    # Reset logic
    if base.Node:
      mt = ctr.createTarget()
      irobot = mt.getSimWorldToRobotWorld()
      irobot.invert()
      pos = irobot * base.Node.WorldPositionMatrix * base.PositionMatrix
      base.Node = None
      base.PositionMatrix = pos
    
    tracking_on = False
    comp.OnNodeTransform = None
    if base_node:
      base_node.OnNodeTransform = None
    base_node = None
    
  #endif
  pass


def format_expression(exp, exe = None):
  # Take exression from prop_form and format it to prop_to
  # Formatted expression supports signal maps and alternative operators etc.
  
  replace_these = {}
  replace_these['ON'] = '1'
  replace_these['OFF'] = '0'
  replace_these['<>'] = '!='
  replace_these['NOT'] = '!'
  replace_these['OR'] = '|'
  replace_these['AND'] = '&'
  
  # Format expression for VC style
  exp2 = exp
  for key, value in replace_these.iteritems():
    exp2 = exp2.replace(key, value)
  
  if not exe:
    return exp2
  
  # If executor is given also replace signal kayewords to current signal values
  
  #DO[x]
  re_exp = '(OUT|DO)\[(%s?)\]' % (RE.ANY)
  exp3 = exp2
  exp4 = ''
  match = re.search(re_exp, exp3, RE.FLAGS)
  while match:
    exp4 += exp3[0:match.start()]
    exp3 = exp3[match.end():]
    #
    index = -1
    try:
      index = int(match.group(2))
    except:
      #Try DO[i] where i is property
      try:
        index = int(comp.getProperty(match.group(2)).Value)
      except:
        print 'Parse error in WaitStatement.'
        return
    if index < 0 or index >= exe.DigitalOutputSignals.PortCount:
      print 'DO index error in WaitStatement.'
      return
    value = exe.DigitalOutputSignals.input(index)
    if value:
      exp4 += '1'
    else:
      exp4 += '0'
    #
    match = re.search(re_exp, exp3, RE.FLAGS)
  exp4 += exp3
  
  #DI[x]
  re_exp = '(IN|DI)\[(%s?)\]' % (RE.ANY)
  exp3 = exp4
  exp4 = ''
  match = re.search(re_exp, exp3, RE.FLAGS)
  while match:
    exp4 += exp3[0:match.start()]
    exp3 = exp3[match.end():]
    #
    index = -1
    try:
      index = int(match.group(2))
    except:
      #Try DO[i] where i is property
      try:
        index = int(comp.getProperty(match.group(2)).Value)
      except:
        print 'Parse error in WaitStatement.'
        return
    if index < 0 or index >= exe.DigitalInputSignals.PortCount:
      print 'DI index error in WaitStatement.'
      return
    value = exe.DigitalInputSignals.input(index)
    if value:
      exp4 += '1'
    else:
      exp4 += '0'
    #
    match = re.search(re_exp, exp3, RE.FLAGS)
  exp4 += exp3
  parsed_exp = exp4
  
  return parsed_exp


def track(arg):
  global ctr, mt, tracking_on, base_node, last_time
  
  # Make sure that sim clock proceeds
  if sim.SimTime <= last_time:
    return
  last_time = sim.SimTime
  
  # Update robot pose
  if ctr and mt and tracking_on:
    mt.Target = mt.Target
    if base_node:
      mt.BaseName = mt.BaseName
    ctr.moveImmediate(mt)